unit Sockets;

{ Install this component using Component, Install, Add within
  Delphi 2.0. This component high-levels the winsock functionallity
  and provides a simple interface to TCP/IP socket functions.

  The code herein is released to the public domain without conditions.

  Written By:      Gary T. Desrosiers
  Date:            March 27th, 1995.
  Modified:        March 18th, 1996 for Delphi 2.0
  Modified:        July 11th, 1997 for Delphi 3.0 by
                   Atul Vaidya, 100025.3466
  Copyright:       (R) Copyright by Gary T. Desrosiers, 1995. All Rights Reserved
  UserID(s):       71062,2754
                   desrosi@pcnet.com

  Description:     This control performs WinSock TCP/IP functions.

  Prerequisites:   You must have the TCP/IP protocol installed and the
                   WSOCK32.DLL available. This code has been tested under
                   Windows 95 and Windows NT 4.0.


  Modifications:   Version 3 - March 18th, 1996
                   - Added properties;
                     - HostName, Returns the name of the local host.
                     - MaximumReceiveLength, Sets the maximum receive buffer size.
                   - Added Methods;
                     - GetLocalIPAddr, returns the IP address of the local host
                   - Added Events;
                     - OnDataNeeded, called when the socket needs data and it's
                        okay to write.

                   Version 2 - July 5th, 1995
                   - Added properties;
                     - MasterSocket, Gets the listener's socket
                     - Peek, Preview data in the input buffer.
                     - NonBlocking, Blocking vs Non-Blocking sockets
                     - Timeout, For blocking mode timeouts
                     - OOB, Sends and receives data out of band (urgent data)
                   - Modified properties;
                     - SocketNumber to read/write
                     - Text (no longer published)
                   - Added Methods;
                     - SCancelListen, new method cancels the listener socket
                     - GetPeerIPAddr, returns partners IP address
                     - GetPeerPort, returns partners port
                   - Modified Methods;
                     - GetIPAddr, Documented and bug fix
                     - GetPort, Documented
                     - SClose, Added shutdown, etc.
                     - SReceive, Modified to use PChar instead of Pascal strings
                     - SSend, Modified to use PChar instead of Pascal strings
                     - SetText, Now loops until entire buffer sent
                   - Added Events
                     - OnErrorOccurred, Called on WinSock errors.

  Properties:       Authorized, true for authorized port assignment
                    default is false which allocates ports > 1024

                    IPAddr, Design time and runtime read/write.
                     Sets the IP Address of the partner that you will
                     eventually SConnect to. You may specify this as
                     dotted decimal or a literal name to be converted
                     via DNS.
                     examples;
                       Sockets1.IPAddr := 'desrosi';
                       Sockets1.IPAddr := '127.0.0.1';
                       addr := Sockets1.IPAddr;

                   Port, Design time and runtime read/write.
                     Sets the Port number of the remote port to connect
                     to or the local port to listen on depending on
                     whether you subsequently issue a SConnect or SListen.
                     This can be specified as a number or a literal name
                     to be converted via DNS.
                     examples;
                       Sockets1.Port := 'echo';
                       Sockets1.Port := '7';
                       port := Sockets1.Port;

                   SocketNumber, Runtime Read/write.
                     Returns (or sets) the socket number of the currently
                     allocated connection.
                     example;
                       sock := Sockets1.SocketNumber;

                   MasterSocket, Runtime Read/Write.
                     Returns (or sets) the master socket number (listener)
                     example;
                       msock := Sockets1.MasterSocket;

                   Text, Design time and runtime read/write.
                     if set, sends the text to the partner.
                     if read, receives some text from the partner.
                     examples;
                       buffer := Sockets1.Text; (* Receive data *)
                       Sockets1.Text := 'This is a test'; (* Send Data *)

                   Peek, runtime read only.
                     Returns up to 255 characters of data waiting to
                     be received but does not actually receive the
                     data.

                   OOB, runtime read/write.
                     if set, sends the text to the partner as urgent (out of
                       band) data.
                     if read, receives urgent (out of band) data.
                     examples;
                       buffer := Sockets1.OOB;
                       Sockets1.OOB := 'This is a test';

                   NonBlocking, Design time and runtime read/write
                     Set to False for blocking mode and True for non-blocking
                     mode (the default). When the socket is in blocking
                     mode, none of the event callback functions (with the
                     exception of OnErrorOccurred) will function.

                   Timeout, Design time and runtime read/write
                     When NonBlocking = 0 (blocking mode) this value
                     specifies the maximum amount of time that
                     a socket operation can take. After this time
                     limit expires, the operation is canceled and
                     an error occurs. The default is 30 (seconds).
                     The Valid range is 0-60 seconds. Setting Timeout
                     to zero causes the operation to wait indefinitely.

                   MaximumReceiveLength, Runtime read/write
                     Set to limit the size of buffers retrieved using
                     .Text, .PeekData, and .OOB Default is 8192.

                   HostName, Runtime read only.
                     Returns the name of the local host.

Methods:           SConnect - Connects to the remote (or local) system
                     specified in the IPAddr and Port properties.
                     example;
                       Sockets1.SConnect; (* Connect to partner *)

                   SListen - Listens on the port specified in the Port
                     property.
                     example;
                       Sockets1.SListen; (* Establish server environment *)

                   SCancelListen - Cancels listens on the socket.
                     example;
                       Sockets1.SCancelListen; (* Dont accept further clients *)

                   SAccept - Accepts a client request. Usually issued in
                     OnSessionAvailable event.
                     example;
                       Sock := Sockets1.SAccept; (* Get client connection *)

                   SClose - Closes the socket.
                     example;
                       Sockets1.SClose; (* Close connection *)

                   SReceive - Receives data from partner, similar to
                     reading the property Text although this function
                     uses PChar instead of Pascal strings.
                     example;
                       len := Sockets1.SReceive(Sockets1.SocketNumber,szBuffer,4096);

                   SSend - Sends data to the partner, similar to
                     setting the property Text although this function
                     uses PChar instead of Pascal strings.
                     example;
                       len := Sockets1.SSend(Sockets1.SocketNumber,szBuff,32000);

                   GetPort - Returns the actual port number of the socket
                     specified as the argument. Generally used when you've
                     specified a port of zero and need to retrieve the
                     assigned port number.

                   GetIPAddr - Returns the IP Address of the socket specified
                     as the argument.

                   GetPeerPort - Returns the partners port number of the socket
                     specified as the argument.

                   GetPeerIPAddr - Returns partners IP Address of the socket
                     specified as the argument.

                   GetLocalIPAddr - Returns the local host's IP Address.

Events:            OnDataAvailable - Called when data is available to
                     be received from the partner. You should issue;
                     buffer := Sockets1.Text; or a SReceive method to
                     receive the data from the partner.

                   OnDataNeeded - Called when it is okay to write
                     data to the socket.

                   OnSessionAvailable - Called when a client has requested
                     to connect to a 'listening' server. You can call
                     the method SAccept here.

                   OnSessionClosed - Called when the partner has closed
                     a socket on you. Normally, you would close your side
                     of the socket when this event happens.

                   OnSessionConnected - Called when the SConnect has
                     completed and the session is connected. This is a
                     good place to send the initial data of a conversation.
                     Also, you may want to enable certain controls that
                     allow the user to send data on the conversation here.

                   OnErrorOccurred - Called when an error occurs on the socket.
                     If defined, the OnErrorOccurred procedure is called when
                     the error occurs. If the procedure isn't defined then
                     a dialog box is displayed with the error text and the
                     program is halted.
}

interface

uses
  SysUtils,WinTypes,WinProcs,Messages,Classes,Graphics,Controls,
  Forms,Dialogs,WinSock;
const
  { User Windows Messages }
  WM_ASYNCSELECT = WM_USER + 0;

type ESocketError = class(Exception);

type
  TDataAvailable=procedure(Sender:TObject;Socket:TSocket) of object;
  TDataNeeded=procedure(Sender:TObject;Socket:TSocket) of object;
  TSessionClosed=procedure(Sender:TObject;Socket:TSocket) of object;
  TSessionAvailable=procedure (Sender:TObject;Socket:TSocket) of object;
  TSessionConnected=procedure (Sender:TObject;Socket:TSocket) of object;
  TErrorOccurred=procedure (Sender:TObject;Socket:Integer;Error:Integer;Msg:string) of object;

  TSockets = class(TWinControl)
  private
    Pse:PServEnt;
    Phe:PHostEnt;
    Ppe:PProtoEnt;
    sin:TSockAddrIn;
    initdata:TWSAData;
    FAuthorized:Boolean;
    FPort:String;
    FIPAddr:String;
    FSocket:TSocket;
    FMSocket:TSocket;
    FMode:longint;
    FTimeout:Integer;
    FMaximumReceiveLength:Integer;
    FDataAvailable:TDataAvailable;
    FDataNeeded:TDataNeeded;
    FSessionClosed:TSessionClosed;
    FSessionAvailable:TSessionAvailable;
    FSessionConnected:TSessionConnected;
    FErrorOccurred:TErrorOccurred;
    procedure SetText(Text:string);
    function GetText:string;
    procedure SetTextOOB(Text:string);
    function GetTextOOB:string;
    function PeekData:string;
    function SocketErrorDesc(error:Integer):string;
    function SocketError(Socket:TSocket;sockfunc:string;error:Integer):Boolean;
    procedure TWMPaint(var msg:TWMPaint); message WM_PAINT;
    procedure SetTimeout;
    procedure ResetTimeout;
    function PortIPCheck:Boolean;
    function SetService:Boolean;
    function MakeAndBind:Boolean;
  protected
    procedure WMASyncSelect(var msg:TMessage);message WM_ASYNCSELECT;
    procedure WMTimer(var msg:TMessage);message WM_TIMER;
  public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
    { I'd like to call these methods Connect, Close, Listen, etc but
      they would conflict with the WSock32.DLL function names ! }
    procedure SConnect;
    procedure SClose;
    procedure SListen;
    procedure SCancelListen;
    function SAccept:TSocket;
    function SReceive(Socket:TSocket;szBuff:PChar;var rlen:Integer):Integer;
    function SSend(Socket: TSocket;szBuff:PChar;var slen:Integer):Integer;
    function GetIPAddr(aSocket:TSocket):string;
    function GetLocalHostName:string;
    function GetLocalIPAddr:string;
    function GetPort(aSocket:TSocket):string;
    function GetPeerIPAddr(aSocket:TSocket):string;
    function GetPeerPort(aSocket:TSocket):string;
    function GetBlocking:Boolean;
    procedure SetBlocking(flag:Boolean);
    property Text:string read GetText write SetText;
    property Authorized:Boolean read FAuthorized write FAuthorized;
    property Peek:string read PeekData;
    property OOB:string read GetTextOOB write SetTextOOB;
    property SocketNumber:TSocket read FSocket write FSocket;
    property MasterSocket:TSocket read FMSocket write FMSocket;
    property HostName:string read GetLocalHostName;
  published
    property MaximumReceiveLength:Integer read FMaximumReceiveLength write FMaximumReceiveLength;
    property IPAddr:string read FIPAddr write FIPAddr;
    property Port:string read FPort write FPort;
    property NonBlocking:Boolean read GetBlocking write SetBlocking default True;
    property Timeout:Integer read FTimeout write FTimeout default 30;
    property OnDataAvailable:TDataAvailable read FDataAvailable
      write FDataAvailable;
    property OnDataNeeded:TDataNeeded read FDataNeeded
      write FDataNeeded;
    property OnSessionClosed:TSessionClosed read FSessionClosed
      write FSessionClosed;
    property OnSessionAvailable:TSessionAvailable read FSessionAvailable
      write FSessionAvailable;
    property OnSessionConnected:TSessionConnected read FSessionConnected
      write FSessionConnected;
    property OnErrorOccurred:TErrorOccurred read FErrorOccurred
      write FErrorOccurred;
  end;

procedure Register;

implementation


procedure Register;
begin
  RegisterComponents('Internet',[TSockets]);
end;


constructor TSockets.Create(AOwner:TComponent);
var iStatus: Integer;
    Mes:String;
begin
  inherited Create(AOwner);
  FAuthorized:=False;
  FMode:=1;
  FTimeout:=15;
  FMaximumReceiveLength:=8192;
  FSocket:=INVALID_SOCKET;
  FMSocket:=INVALID_SOCKET;
  iStatus:= WSAStartup($101,initdata);
  case iStatus of
    WSASYSNOTREADY:Mes:='Network subsystem not ready.';
    WSAVERNOTSUPPORTED:Mes:='Requested socket version not provided.';
    WSAEINVAL:Mes:='WinSock does not support socket version.';
    0:Mes:='Socket initialized successfully.';
    else Mes:='Error ' + IntToStr(iStatus) + ' during initialization.';
  end;
  ShowMessage(Mes);
  invalidate;
end;

destructor TSockets.Destroy;
var iStatus:Integer;
begin
  iStatus:=WSACleanup;
  if iStatus < 0 then
    SocketError(INVALID_SOCKET,'Destructor (WSACleanup)',WSAGetLastError);
  inherited Destroy;
end;

procedure TSockets.TWMPaint(var msg: TWMPaint);
var icon:HIcon;
    dc:HDC;
begin
  if csDesigning in ComponentState then
  begin
    icon:=LoadIcon(HInstance,MAKEINTRESOURCE('TSOCKETS'));
    dc:=GetDC(Handle);
    Width:=15;
    Height:=15;
    DrawIcon(dc,0,0,icon);
    ReleaseDC(Handle,dc);
    FreeResource(icon);
  end;
  ValidateRect(Handle,nil);
end;

function TSockets.GetBlocking: Boolean;
begin
  Result:=(FMode = 1);
end;

procedure TSockets.SetBlocking(flag: Boolean);
begin
  FMode:=Integer(Flag);
end;

procedure TSockets.SetText(Text:string);
var BytesSent:Integer;
    pBuf:String;
begin
  pBuf:=Text;
  if FMode = 0 then SetTimeout;
  BytesSent:=send(FSocket,pBuf[1],Length(Text),0);
  if FMode = 0 then ResetTimeout;
  if BytesSent < 0 then
    SocketError(FSocket,'SetText (Send)',WSAGetLastError);
end;

function TSockets.GetText:string;
var len:Integer;
    pBuf:String;
begin
  Result:='';
  if FSocket = INVALID_SOCKET then exit;
  if FMode = 0 then SetTimeout;
  SetLength(pBuf,FMaximumReceiveLength);
  len:=recv(FSocket,pBuf[1],FMaximumReceiveLength,0);
  if FMode = 0 then ResetTimeout;
  if len < 0 then
  SocketError(FSocket,'GetText (Recv)',WSAGetLastError);
end;

procedure TSockets.SetTextOOB(Text:String);
var BytesLeft,BytesSent:Integer;
    pBuf:String;
begin
  pBuf:=Text;
  if FMode = 0 then SetTimeout;
  BytesSent:=send(FSocket,pBuf[1],Length(Text),MSG_OOB);
  if FMode = 0 then ResetTimeout;
  if BytesSent < 0 then
    SocketError(FSocket,'SetText (Send)',WSAGetLastError);
end;

function TSockets.GetTextOOB:String;
var len:Integer;
    pBuf:String;
begin
  Result:='';
  if FSocket = INVALID_SOCKET then exit;
  SetLength(pBuf,FMaximumReceiveLength);
  if FMode = 0 then SetTimeout;
  len:=recv(FSocket,pBuf[1],FMaximumReceiveLength,MSG_OOB);
  if FMode = 0 then ResetTimeout;
  if len < 0 then
      SocketError(FSocket,'GetText (Recv)',WSAGetLastError);
  Result:=pBuf;
end;

function TSockets.PeekData:string;
var len:Integer;
    pBuf:String;
begin
  Result:='';
  if FSocket = INVALID_SOCKET then exit;
  SetLength(pBuf,FMaximumReceiveLength);
  if FMode = 0 then SetTimeout;
  len:=recv(FSocket,pBuf[1],FMaximumReceiveLength,MSG_PEEK);
  if FMode = 0 then ResetTimeout;
  if len < 0 then
      SocketError(FSocket,'PeekData (Peek)',WSAGetLastError);
  Result:=pBuf;
end;

function TSockets.GetPort(aSocket: TSocket): string;
var addr:TSockAddrIn;
    addrlen:Integer;
begin
  addrlen:=sizeof(addr);
  getsockname(aSocket,addr,addrlen);
  Result:=IntToStr(ntohs(addr.sin_port));
end;

function TSockets.GetIPAddr(aSocket:TSocket):string;
var addr:TSockAddrIn;
    addrlen:Integer;
begin
  addrlen:=sizeof(addr);
  getsockname(aSocket,addr,addrlen);
  Result:=inet_ntoa(addr.sin_addr);
end;

function TSockets.GetLocalIPAddr:string;
var addr:TSockAddrIn;
    Phe:PHostEnt;
    szHostName:String;
begin
  try
    SetLength(szHostName,128);
    GetHostName(PChar(szHostName),128);
    Phe:=GetHostByName(PChar(szHostName));
    if Phe = nil then Result:='0.0.0.0' else
    begin
      addr.sin_addr.S_addr:=longint(plongint(Phe^.h_addr_list^)^);
      Result:=inet_ntoa(addr.sin_addr);
    end;
    except Result:='0.0.0.0' end;
end;

function TSockets.GetLocalHostName:string;
var szHostName:String;
begin
  try
    SetLength(Result,128);
    GetHostName(PChar(Result),128);
  except Result:='' end;
end;

function TSockets.GetPeerPort(aSocket:TSocket):string;
var addr: TSockAddrIn;
    addrlen: Integer;
begin
  try
    addrlen:=sizeof(addr);
    getpeername(aSocket,addr,addrlen);
    Result:=IntToStr(ntohs(addr.sin_port));
  except Result:='-1' end;
end;

function TSockets.GetPeerIPAddr(aSocket:TSocket):string;
var addr: TSockAddrIn;
    addrlen: Integer;
begin
  try
    addrlen:=sizeof(addr);
    getpeername(aSocket,addr,addrlen);
    Result:=String(inet_ntoa(addr.sin_addr));
  except Result:='' end;
end;

function TSockets.SReceive(Socket:TSocket;szBuff:PChar;var rlen:Integer):Integer;
begin
  Result:=-1;
  try
    if Socket = INVALID_SOCKET then exit;
    if FMode = 0 then SetTimeout;
    Result:=recv(Socket,szBuff^,rlen,0);
    if FMode = 0 then ResetTimeout;
    if rlen < 0 then
    SocketError(FSocket,'SReceive',WSAGetLastError);
  except end;
end;

function TSockets.SSend(Socket:TSocket;szBuff:PChar;var slen:Integer):Integer;
begin
  if Socket = INVALID_SOCKET then exit;
  try
    if FMode = 0 then SetTimeout;
    slen:=send(Socket,szBuff^,slen,0);
    if FMode = 0 then ResetTimeout;
    if slen < 0 then SocketError(FSocket,'SSend',WSAGetLastError);
    Result:=slen;
  except Result:=-1 end;
end;

procedure TSockets.WMASyncSelect(var msg:TMessage);
var err:Integer;
    errfn:string;
begin
  err:=WSAGetSelectError(msg.LParam);
  if err > WSABASEERR then
  begin
    case WSAGetSelectEvent(msg.lParam) of
      FD_READ:errfn:='FD_READ';
      FD_WRITE:errfn:='FD_WRITE';
      FD_CLOSE:errfn:='FD_CLOSE';
      FD_ACCEPT:errfn:='FD_ACCEPT';
      FD_CONNECT:errfn:='FD_CONNECT';
    end;
    SocketError(msg.wParam,errfn,err);
  end
  else
  case WSAGetSelectEvent(msg.lParam) of
    FD_READ:if Assigned(FDataAvailable) then FDataAvailable(Self,msg.wParam);
    FD_WRITE:if Assigned(FDataNeeded) then FDataNeeded(Self,msg.wParam);
    FD_CLOSE:if Assigned(FSessionClosed) then FSessionClosed(Self,msg.wParam);
    FD_ACCEPT:if Assigned(FSessionAvailable) then FSessionAvailable(Self,msg.wParam);
    FD_CONNECT:if Assigned(FSessionConnected) then FSessionConnected(Self,msg.wParam);
  end;
end;

procedure TSockets.WMTimer(var msg:TMessage);
begin
  KillTimer(Handle,10);
  if WSAIsBlocking then
  begin
    WSACancelBlockingCall;
    if Assigned(FErrorOccurred) then
      FErrorOccurred(Self,FSocket,WSAETIMEDOUT,'Blocking call timed out')
    else ShowMessage('Blocking call timed out.');
  end;
end;

function TSockets.PortIPCheck:Boolean;
begin
  try
   if FPort = '' then
   raise ESocketError.Create('No port specified.');
   if FIPAddr = '' then
   raise ESocketError.Create('No IP address specified.');
   Result:=True;
  except Result:=False end;
end;

function TSockets.SetService:Boolean;
begin
  try
    sin.sin_family:=AF_INET;
    Pse:=getservbyname(PChar(FPort),PChar('TCP'));
    if Pse = nil then sin.sin_port:=htons(StrToInt(FPort))
    else sin.sin_port:=Pse^.s_port;
    sin.sin_addr.s_addr:=inet_addr(PChar(FIPAddr));
    if sin.sin_addr.s_addr = INADDR_NONE then
    begin
    Phe:=gethostbyname(PChar(FIPAddr));
    if Phe = nil then raise
        ESocketError.Create('Cannot convert host address.');
        sin.sin_addr.S_addr:=longint(plongint(Phe^.h_addr_list^)^);
    end;
    Result:=True;
 except Result:=False end;
end;

function TSockets.MakeAndBind:Boolean;
var alPort:TSocket;
    bind_sin:TSockAddrIn;
begin
  try
    Result:=True;
    Ppe:=nil;
    Ppe:=getprotobyname(PChar('TCP'));
    FSocket:=socket(PF_INET,SOCK_STREAM,Ppe^.p_proto);
    if FSocket = INVALID_SOCKET then raise
       ESocketError.Create('Invalid socket.');
    if FAuthorized = True then
    begin
      alport:=IPPORT_RESERVED;
      bind_sin.sin_family:=AF_INET;
      bind_sin.sin_addr.s_addr:=0;
      repeat
         bind_sin.sin_port:=htons(alport);
         if bind(FSocket,bind_sin,sizeof(bind_sin)) = 0 then
         break;
         if WSAGetLastError <> WSAEADDRINUSE then
         Result:=SocketError(FSocket,'SConnect bind()',WSAGetLastError);
         dec(alport);
      until(alport <= (IPPORT_RESERVED div 2)) or not(Result);
      if Result then Result:=(alPort > IPPORT_RESERVED div 2);
  end;
  except Result:=False end;
end;

procedure TSockets.SConnect;
var
  iStatus: Integer;
begin
  if not(PortIPCheck) then exit;
  SetService;
  if not(MakeAndBind) then exit;
  if FMode = 1 then
  begin
    iStatus:=WSAASyncSelect(FSocket,Handle,WM_ASYNCSELECT,
      FD_READ or FD_CLOSE or FD_CONNECT or FD_WRITE);
    if iStatus <> 0 then
      SocketError(FSocket,'WSAAsyncSelect',WSAGetLastError);
  end else iStatus:=ioctlsocket(FSocket,FIONBIO,FMode);
  if FMode = 0 then SetTimeout;
  iStatus:=connect(FSocket,sin,sizeof(sin));
  if FMode = 0 then ResetTimeout;
  if iStatus <> 0 then
  begin
    iStatus:=WSAGetLastError;
    if iStatus <> WSAEWOULDBLOCK then
       SocketError(FSocket,'SConnect',WSAGetLastError);
  end;
end;

procedure TSockets.SListen;
var iStatus:Integer;
begin
  if not(PortIPCheck) then exit;
  sin.sin_family:=AF_INET;
  sin.sin_addr.s_addr:=INADDR_ANY;
  Pse:=getservbyname(PChar(FPort),PChar('tcp'));
  if Pse = nil then
     sin.sin_port:=htons(StrToInt(FPort))
  else sin.sin_port:=Pse^.s_port;
  Ppe:=getprotobyname(PChar('tcp'));
  FMSocket:=socket(PF_INET,SOCK_STREAM,Ppe^.p_proto);
  if FMSocket < 0 then
    SocketError(INVALID_SOCKET,'socket',WSAGetLastError);
  iStatus:=bind(FMSocket, sin, sizeof(sin));
  if iStatus <> 0 then SocketError(FMSocket,'Bind',WSAGetLastError);
  iStatus:=listen(FMSocket,5);
  if iStatus <> 0 then SocketError(FMSocket,'Listen',WSAGetLastError);
  if FMode = 1 then
  begin
    iStatus:=WSAASyncSelect(FMSocket,Handle,WM_ASYNCSELECT,
      FD_READ or FD_WRITE or FD_ACCEPT or FD_CLOSE);
    if iStatus <> 0 then
      SocketError(FMSocket,'WSAASyncSelect',WSAGetLastError);
  end else ioctlsocket(FMSocket,FIONBIO,FMode);
end;

procedure TSockets.SCancelListen;
var iStatus:Integer;
begin
  if FMode = 1 then
    WSAASyncSelect(FMSocket,Handle,WM_ASYNCSELECT,0);
  shutdown(FMSocket,2);
  iStatus:=closesocket(FMSocket);
  if iStatus <> 0 then
    SocketError(FMSocket,'CancelListen (closesocket)',WSAGetLastError);
  FMSocket:=0;
end;

function TSockets.SAccept:TSocket;
var iStatus:Integer;
    len:Integer;
begin
  len:=sizeof(sin);
  if FMode = 0 then SetTimeout;
  FSocket:=accept(FMSocket,@sin,@len);
  if FMode = 0 then
  begin
    ResetTimeout;
    ioctlsocket(FSocket,FIONBIO,FMode);
  end;
  if FMSocket < 0 then
    SocketError(FSocket,'Accept',WSAGetLastError);
  Result:=FSocket;
end;

procedure TSockets.SClose;
var iStatus: Integer;
    lin: TLinger;
    linx: array[0..3] of char absolute lin;
begin
  if FMode = 1 then WSAASyncSelect(FSocket,Handle,WM_ASYNCSELECT,0);
  if WSAIsBlocking then WSACancelBlockingCall;
  shutdown(FSocket,2);
  lin.l_onoff:=1;
  lin.l_linger:=0;
  setsockopt(FSocket,SOL_SOCKET,SO_LINGER,linx,sizeof(lin));
  iStatus:=closesocket(FSocket);
  if iStatus <> 0 then
    SocketError(FSocket,'Disconnect (closesocket)',WSAGetLastError);
  FSocket:=INVALID_SOCKET;
end;

function TSockets.SocketError(Socket:TSocket;sockfunc:string;error:Integer):Boolean;
var ErrMsg:string;
begin
  Result:=True;
  ErrMsg:=SocketErrorDesc(error);
  ErrMsg:='Error '+IntToStr(error) + ' in function ' + sockfunc +
  #13#10 + ErrMsg;
  if Assigned(FErrorOccurred) then FErrorOccurred(Self,Socket,error,ErrMsg)
  else
  begin
    ShowMessage(ErrMsg);
    Result:=False;
  end;
end;

function TSockets.SocketErrorDesc(error:Integer):string;
begin
  case error of
    WSAEINTR:Result:='Interrupted system call';
    WSAEBADF:Result:='Bad file number';
    WSAEACCES:Result:='Permission denied';
    WSAEFAULT:Result:='Bad address';
    WSAEINVAL:Result:='Invalid argument';
    WSAEMFILE:Result:='Too many open files';
    WSAEWOULDBLOCK:Result:='Operation would block';
    WSAEINPROGRESS:Result:='Operation now in progress';
    WSAEALREADY:Result:='Operation already in progress';
    WSAENOTSOCK:Result:='Socket operation on non-socket';
    WSAEDESTADDRREQ:Result:='Destination address required';
    WSAEMSGSIZE:Result:='Message too long';
    WSAEPROTOTYPE:Result:='Protocol wrong type for socket';
    WSAENOPROTOOPT:Result:='Protocol not available';
    WSAEPROTONOSUPPORT:Result:='Protocol not supported';
    WSAESOCKTNOSUPPORT:Result:='Socket type not supported';
    WSAEOPNOTSUPP:Result:='Operation not supported on socket';
    WSAEPFNOSUPPORT:Result:='Protocol family not supported';
    WSAEAFNOSUPPORT:Result:='Address family not supported by protocol family';
    WSAEADDRINUSE:Result:='Address already in use';
    WSAEADDRNOTAVAIL:Result:='Can''t assign requested address';
    WSAENETDOWN:Result:='Network is down';
    WSAENETUNREACH:Result:='Network is unreachable';
    WSAENETRESET:Result:='Network dropped connection on reset';
    WSAECONNABORTED:Result:='Software caused connection abort';
    WSAECONNRESET:Result:='Connection reset by peer';
    WSAENOBUFS:Result:='No buffer space available';
    WSAEISCONN:Result:='Socket is already connected';
    WSAENOTCONN:Result:='Socket is not connected';
    WSAESHUTDOWN:Result:='Can''t send after socket shutdown';
    WSAETOOMANYREFS:Result:='Too many references: can''t splice';
    WSAETIMEDOUT:Result:='Connection timed out';
    WSAECONNREFUSED:Result:='Connection refused';
    WSAELOOP:Result:='Too many levels of symbolic links';
    WSAENAMETOOLONG:Result:='File name too long';
    WSAEHOSTDOWN:Result:='Host is down';
    WSAEHOSTUNREACH:Result:='No route to host';
    WSAENOTEMPTY:Result:='Directory not empty';
    WSAEPROCLIM:Result:='Too many processes';
    WSAEUSERS:Result:='Too many users';
    WSAEDQUOT:Result:='Disc quota exceeded';
    WSAESTALE:Result:='Stale NFS file handle';
    WSAEREMOTE:Result:='Too many levels of remote in path';
    WSASYSNOTREADY:Result:='Network sub-system is unusable';
    WSAVERNOTSUPPORTED:Result:='WinSock DLL cannot support this application';
    WSANOTINITIALISED:Result:='WinSock not initialized';
    WSAHOST_NOT_FOUND:Result:='Host not found';
    WSATRY_AGAIN:Result:='Non-authoritative host not found';
    WSANO_RECOVERY:Result:='Non-recoverable error';
    WSANO_DATA:Result:='No Data';
    else Result:='Not a WinSock error';
  end;
end;

procedure TSockets.SetTimeout;
begin
  if FTimeout > 0 then
    SetTimer(Handle,10,FTimeout*1000,nil);
end;

procedure TSockets.ResetTimeout;
begin
  if FTimeout > 0 then
    KillTimer(Handle,10);
end;

end.
